# Generated by Django 3.0.6 on 2020-05-22 15:04

from django.db import migrations
import djmoney.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0016_longer_product_code'),
    ]

    operations = [
        migrations.AlterField(
            model_name='charge',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, max_digits=12),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, max_digits=12),
        ),
    ]
